
Object subclass: #DifferentialEquationSolver
  instanceVariableNames: 
    ' differentiationBlock successes successesBeforeIncreasingH solution0 solution1 solution2 k1 k2 k3 k4 y y1 y2 y3 dydt dydt0 dydt1 dydt2 dydt3 h minimumH relativeAccuracy '
  classVariableNames: ''
  poolDictionaries: ''
  category: 'JOOP10'
!


!DifferentialEquationSolver class methodsFor: 'instance creation'!
   
solverFor: aBlock
	^self new differentiationBlock: aBlock
!
   
new
	^super new initialize
! !

!DifferentialEquationSolver class methodsFor: 'comment'!

comment
	"K * hWanted **5 /K hUsed **5 = eWanted / eUsed
	So hWanted = hUsed * (eWanted / eUsed) ** (1/5).
	i.e. hNew = h * (eWanted / eUsed) ** (1/5)."
! !

!DifferentialEquationSolver class methodsFor: 'examples'!
   
example3
	"DifferentialEquationSolver example3"
	| solver t0 y0 solution1 solution2 |
	solver := self example1.
	t0 := 0.0. y0 := #(10.0 0.0).
	Transcript cr; cr; show: 'Time: Exact => Computed'.
	Transcript cr; show: t0 printString; show: ': '; show: y0 printString; space; show: y0 printString.
	1.0 to: 10.0 by: 1.0 do: [:t |
		solution1 := self exactSolution1At: t given: y0 at: t0.
		solution2 := self approximateSolution1At: t given: y0 at: t0 and: solver.
		t0 := t. y0 := solution2.
		Transcript cr; show: t printString; show: ': ['.
		Transcript show: solution1 first terseString; space; show: solution1 last terseString; show: '], ['.
		Transcript show: solution2 first terseString; space; show: solution2 last terseString; show: ']'.
		"Transcript cr; show: t printString; show: ': '; show: solution1 printString; space; show: solution2 printString"]
!
  
example: exampleNumber solveFrom: t0 to: t1 by: step initially: y0
	"DifferentialEquationSolver example: 1 solveFrom: 0.0 to: 10.0 by: 1.0 initially: #(0.0).
	DifferentialEquationSolver example: 2 solveFrom: 0.0 to: 10.0 by: 1.0 initially: #(10.0 10.0)"
	| which solverName exactSolutionMethod solver t y solution1 solution2 |
	which := exampleNumber printString.
	solverName := ('solver', which) asSymbol.
	exactSolutionMethod := ('exactSolution', which, 'At:given:at:') asSymbol.

	self printTitle: 'Solving example ', which.
	solver := self perform: solverName.

	t := t0. y := y0.
	self printTime: t solutions: y and: y.

	"The exact solution is computed from y0 and t0 whereas
	the approximate solution is computed from the last y."
	t0 + step to: t1 by: step do: [:newT |
		solution1 := self perform: exactSolutionMethod with: newT with: y0 with: t0.
		solution2 := solver solveAt: newT given: y at: t.
		self printTime: newT solutions: solution1 and: solution2.
		t := newT. y := solution2].

	Transcript cr; show: 'All done'
!
   
example1
	"DifferentialEquationSolver example1
		solveAt: 1.0 given: #(10.0 0.0) at: 0.0"
	| a b |
	a := 2.0. b := 3.0.
	^self solverFor: [:dydt :t :y |
		"Computes y' and solves for y in y' = f(t,y)..."
		"a = v' = -av where a = 2 whose solution is v = e-at + C1."
		dydt at: 1 put: a negated * y first.
		"v = x' = b where b = 3 whose solution is x = bt + C2."
		dydt at: 2 put: b]
!

example2
	"DifferentialEquationSolver example2"
	| solver t0 y0 solution1 solution2 |
	solver := self example1.
	t0 := 0.0. y0 := #(10.0 0.0).
	Transcript cr; cr; show: 'Time: Exact => Computed'.
	Transcript cr; show: t0 printString; show: ': '; show: y0 printString; space; show: y0 printString.
	1.0 to: 10.0 by: 1.0 do: [:t |
		solution1 := self exactSolution1At: t given: y0 at: t0.
		solution2 := self approximateSolution1At: t given: y0 at: t0 and: solver.
		t0 := t. y0 := solution2.
		Transcript cr; show: t printString; show: ': ['.
		Transcript show: solution1 first terseString; space; show: solution1 last terseString; show: '], ['.
		Transcript show: solution2 first terseString; space; show: solution2 last terseString; show: ']'.
		"Transcript cr; show: t printString; show: ': '; show: solution1 printString; space; show: solution2 printString"]
! !

!DifferentialEquationSolver class methodsFor: 'printing'!
   
printTime: t solutions: y1 and: y2
	Transcript cr; show: t terseString; show: ': '.
	Transcript show: '['; show: y1 first terseString.
	y1 allButFirstDo: [:number |
		Transcript show: ', '; show: number terseString].
	Transcript show: '] => ['; show: y2 first terseString.
	y2 allButFirstDo: [:number |
		Transcript show: ', '; show: number terseString].
	Transcript show: '].'
!

printTitle: aString
	Transcript
		cr; cr; show: aString;
		cr; show: 'Time: Exact => Computed'.
! !

!DifferentialEquationSolver class methodsFor: 'solutions'!
   
solver2
	"DifferentialEquationSolver solver2
		solveAt: 1.0 given: #(10.0 0.0) at: 0.0"
	| a b |
	a := 2.0. b := 3.0.
	^self solverFor: [:dydt :t :y |
		"Computes y' and solves for y in y' = f(t,y)..."
		"a = v' = -av where a = 2 whose solution is v = k1 e-at."
		dydt at: 1 put: a negated * y first.
"Transcript cr; show: 'At t = ', t terseString, ', v'' = -av = -', a printString, '*', y first printString, ' = ', (a negated * y first) terseString."
		"v = x' = b where b = 3 whose solution is x = bt + k2."
		dydt at: 2 put: b]
!

solver1
	"DifferentialEquationSolver solver1 solveAt: 1.0 given: #(0.0) at: 0.0"
	"Computes y'. Used elsewhere to solve for y in y' = f(t,y)..."
	^self solverFor: [:dydt :t :y |
		"x' = 1 whose solution is x = t + k. With initial conditions t=0, x = 0, k = x - t = 0."
		dydt at: 1 put: 1.0]
!
 
exactSolution2At: t given: y0 at: t0
	"Solve for vector y at time t given solution vector y0 at time t0 where
	y is the differential equation of the form y' = f (t, y).

	Let y = <v, x> where problem is
			v' = -av where a = 2,   x' = b where b = 3
	Solution is
			v = k1e-at ,  x = bt + k2"

	| a b k1 k2 |
	a := 2.0. b := 3.0.
	"Compute value of k from initial values."
	k1 := y0 first / (a negated * t0) exp.
	k2 := y0 last - (b * t0).
	^Array
		with: k1 * (a negated * t) exp
		with: (b * t) + k2
!
 
exactSolution1At: t given: y0 at: t0
		"x' = 1 whose solution is x = t + c. With initial conditions t=0, x = 0, c = x - t = 0."
	"Solve for vector y at time t given solution vector y0 at time t0 where
	y is the differential equation of the form y' = f (t, y).

	Let y = <x> where problem is
		x' = 1 whose solution is x = t + k. With initial conditions t=t0, x = y0[1], k = y[0] - t0."

	| k |
	"Compute values of k from initial values."
	k := y0 first - t0.
	^Array with: t + k
! !


!DifferentialEquationSolver methodsFor: 'initializing/finalizing'!
  
reinitializefor: sampleSolution
	y size = sampleSolution size ifTrue: [^self].
	solution0 := sampleSolution shallowCopy.
	solution1 := sampleSolution shallowCopy.
	solution2 := sampleSolution shallowCopy.
	y := sampleSolution shallowCopy.
	k1 := sampleSolution shallowCopy.
	k2 := sampleSolution shallowCopy.
	k3 := sampleSolution shallowCopy.
	k4 := sampleSolution shallowCopy.
	y1 := sampleSolution shallowCopy.
	y2 := sampleSolution shallowCopy.
	y3 := sampleSolution shallowCopy.
	dydt := sampleSolution shallowCopy.
	dydt0 := sampleSolution shallowCopy.
	dydt1 := sampleSolution shallowCopy.
	dydt2 := sampleSolution shallowCopy.
	dydt3 := sampleSolution shallowCopy.

!
  
initialize
	self h: 1.0.
	self minimumH: 1.0e-5.
	self successesBeforeIncreasingH: 3.
	successes := 0.
	self relativeAccuracy: 0.0001.
	self reinitializefor: #()
! !

!DifferentialEquationSolver methodsFor: 'get/set'!
  
relativeAccuracy
	^relativeAccuracy
!
 
successesBeforeIncreasingH: anInteger
	successesBeforeIncreasingH := anInteger
!
  
h: aNumber
	h := aNumber
!

h
	^h
!
   
differentiationBlock
	^differentiationBlock
!
 
relativeAccuracy: aNumber
	relativeAccuracy := aNumber
!
  
minimumH: aNumber
	minimumH := aNumber
!
  
differentiationBlock: aBlock
	differentiationBlock := aBlock
!

successesBeforeIncreasingH
	^successesBeforeIncreasingH
!
 
minimumH
	^minimumH
! !

!DifferentialEquationSolver methodsFor: 'evaluating'!
  
dydtAt: t and: y andSet: result
	self differentiationBlock value: result value: t value: y
! !

!DifferentialEquationSolver methodsFor: 'solving - public'!
 
solveAt: t given: y0 at: t0
	self reinitializefor: y0.
	successes := successes + 1.
	successes > self successesBeforeIncreasingH ifTrue: [
		"With enough successes in a row, we can increase h."
		successes := 0.
		h := 1.9 *h. "a little less than double"
		"Transcript cr; show: 'Increase h to ', h terseString"].
	t > t0
		ifTrue: [^self solveAtIncreasing: t given: y0 at: t0]
		ifFalse: [^self solveAtDecreasing: t given: y0 at: t0]
! !

!DifferentialEquationSolver methodsFor: 'solving - private step 1'!

solveAtDecreasing: t1 given: y0 at: t0
	| t restoreHIfDone oldH |
	restoreHIfDone := false. t := t0. y assign: y0.
	[true] whileTrue: [
		"Avoid overshoot by adjusting h (restore before leaving)."
		t - h < t1 ifTrue: [restoreHIfDone := true. oldH := h. h := t - t1].
		self dydtAt: t and: y andSet: dydt.
		"Next step can decrease h (which means we will go around loop)."
		self step: y and: dydt at: t.
		t := t - h.
		 "Quit if done but restore h if saved."
		t  <= t1 ifTrue: [restoreHIfDone ifTrue: [h := oldH]. ^y].
		restoreHIfDone := false]
!
 
solveAtIncreasing: t1 given: y0 at: t0
	| t restoreHIfDone oldH |
	restoreHIfDone := false. t := t0. y assign: y0.
	[true] whileTrue: [
		"Avoid overshoot by adjusting h (restore before leaving)."
		t + h > t1 ifTrue: [restoreHIfDone := true. oldH := h. h := t1 - t].
		self dydtAt: t and: y andSet: dydt.
		"Next step can decrease h (which means we will go around loop)."
		self step: y and: dydt at: t.
		t := t + h.
"Transcript cr; show: 'So far, solution at ', t terseString, ' is ', y first terseString."
		 "Quit if done but restore h if saved."
		t  >= t1 ifTrue: [restoreHIfDone ifTrue: [h := oldH]. ^y].
		restoreHIfDone := false]
! !

!DifferentialEquationSolver methodsFor: 'solving - private step 2'!

errorIsSmall
	^self computeMaximumRelativeError < self relativeAccuracy
!
 
computeMaximumRelativeError
	| maximumRelativeError absoluteError absoluteValue relativeError save |
	maximumRelativeError := 0.0.
	solution1 with: solution2 do: [:y1 :y2 |
		absoluteError := (y1 - y2) abs.
		"Near 0, use absolute error; otherwise, relative error."
		absoluteValue := (y1 abs max: y2 abs).
		relativeError := absoluteValue < 1.0e-30
			ifTrue: [absoluteError]
			ifFalse: [absoluteError / absoluteValue].
"relativeError > maximumRelativeError ifTrue: [save := absoluteValue]."
		maximumRelativeError := maximumRelativeError max: relativeError].
"Transcript cr; show: 'Error ', maximumRelativeError terseString, ', value ', save terseString."
	^maximumRelativeError
!
   
step: y and: dydt at: t
	"Given a solution for y and dydt at t, find an h beginning with the current
	value or smaller (a minimum exists) for which a solution at t + h is accurate
	to our specification. Approach is to solve twice: (1) once in two half
	steps: t0=t+h/2, then t1 = t0+h/2, (2) once in a full step, t2 = t+h.
	If the t1 and t2 results are close enough, combine them for a fifth order
	Runga-Kutta solution. Otherwise, decrease the h and try again. Notice that
	the t+h value for the solution will depend on the value of h when done.
	A suggestion for a new h is returned when done."
	| halfH t0 error |
	[true] whileTrue: [
		"Take 2 half steps..."
		halfH := 0.5 * h. t0 := t + halfH.
		self into: solution0 rungaKuttaStep: y and: dydt at: t h: halfH.
		self dydtAt: t0 and: solution0 andSet: dydt0.
		self into: solution1 rungaKuttaStep: solution0 and: dydt0 at: t0 h: halfH.
		"Take 1 full step..."
		self into: solution2 rungaKuttaStep: y and: dydt at: t h: h.
		"Compare the two solutions."
		(halfH < self minimumH or: [self errorIsSmall]) ifTrue: [
			y with: solution1 with: solution2 set: [:y1 :y2 |
				"Solution is Runga-Kutta fifth order."
				y1 + ((y1 - y2) / 15.0)].
"Transcript cr; show: 'Solution at ', (t + h) terseString, ' is ', (y first) terseString."
			^self].
		"No, determine next value of h to try."
		h := halfH. successes := 0.
		"Transcript cr; show: 'Try again, decrease h to ', h terseString, ', t = ', t terseString"]
! !

!DifferentialEquationSolver methodsFor: 'solving - private step 3'!
  
into: solution rungaKuttaStep: y0 and: dydt0 at: t0 h: h
	| t1 t2 t3 |
	t1 := t0 + (0.5*h). t2 := t1. t3 := t0 + h.
	"k1 = h*f'(t0,y0)"
	k1 with: dydt0 set: [:_dydt0 | h * _dydt0].
	"k2 = h*f'(t1,y0+k1/2); let y1 = y0+k1/2 (a vector)."
	y1 with: y0 with: k1 set: [:_y0 :_k1 | _y0 + (_k1*0.5)].
	self dydtAt: t1 and: y1 andSet: dydt1.
	k2 with: dydt1 set: [:_dydt1 | h * _dydt1].
	"k3 = h*f'(t2,y0+k2/2); let y2 = y0+k2/2 (a vector)."
	y2 with: y0 with: k2 set: [:_y0 :_k2 | _y0 + (_k2*0.5)].
	self dydtAt: t2 and: y2 andSet: dydt2.
	k3 with: dydt2 set: [:_dydt2 | h * _dydt2].
	"k4 = h*f'(t3,y0+k3); let y3 = y0+k3 (a vector)."
	y3 with: y0 with: k3 set: [:_y0 :_k3 | _y0 + _k3].
	self dydtAt: t3 and: y3 andSet: dydt3.
	k4 with: dydt3 set: [:_dydt3 | h * _dydt3].

	"y = y0 + k1/6 + k2/3 + k3/3 + k4/6."
	solution with: y0 with: k1 with: k2 with: k3 with: k4 set: [:_y0 :_k1 :_k2 :_k3 :_k4 |
		_y0 + ( 0.16666667 * _k1) + ( 0.33333333 * (_k2 + _k3)) + ( 0.16666667 * _k4)]
! !
DifferentialEquationSolver comment: ''!
